<?php
/**
 * @Copyright
 * @package     PCE - Page Cache Extended for Joomla! 3.x
 * @author      Viktor Vogel <admin@kubik-rubik.de>
 * @version     3.3.2 - 2019-07-30
 * @link        https://kubik-rubik.de/pce-page-cache-extended
 *
 * @license     GNU/GPL
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
defined('_JEXEC') || die('Restricted access');

class PlgSystemPageCacheExtended extends JPlugin
{
    protected $app;
    protected $cache;
    protected $cacheKey;
    protected $defaultGroup = 'pce';
    protected $encodingGzip = false;

    public function __construct(&$subject, $config)
    {
        $this->app = JFactory::getApplication();

        if ($this->checkVersionRequirements(false, '3.2', 'PLG_PAGECACHEEXTENDED', 'plg_system_pagecacheextended', JPATH_ADMINISTRATOR)) {
            parent::__construct($subject, $config);

            $this->setDefaultGroupDeviceSpecific();

            $options = array(
                'defaultgroup' => $this->defaultGroup,
                'browsercache' => $this->params->get('browsercache', false),
                'caching'      => false,
            );

            $cacheTime = (int) $this->params->get('cache_time', 15);

            if ($cacheTime > 0) {
                $options['lifetime'] = $cacheTime;
            }

            $this->cache = JCache::getInstance('output', $options);
            $this->cacheKey = JUri::getInstance()->toString();
            $this->checkEncodingGzip();
        }
    }

    /**
     * Checks whether all requirements are met for the execution
     * Written generically to be used in all Kubik-Rubik Joomla! Extensions
     *
     * @param boolean $admin                 Allow backend execution - true or false
     * @param string  $versionMin            Minimum required Joomla! version - e.g. 3.2
     * @param string  $extensionName         Name of the extension of the warning message
     * @param string  $extensionSystemName   System name of the extension for the language file loading - e.g.
     *                                       plg_system_xxx
     * @param string  $jPath                 Path of the language file - JPATH_ADMINISTRATOR or JPATH_SITE
     *
     * @return bool
     * @throws Exception
     */
    private function checkVersionRequirements($admin, $versionMin, $extensionName, $extensionSystemName, $jPath)
    {
        $execution = true;
        $version = new JVersion();

        if (!$version->isCompatible($versionMin)) {
            $execution = false;
            $backendMessage = true;
        }

        if (empty($admin)) {
            if ($this->app->isAdmin()) {
                $execution = false;

                if (!empty($backendMessage)) {
                    $this->loadLanguage($extensionSystemName, $jPath);
                    $this->app->enqueueMessage(JText::sprintf('KR_JOOMLA_VERSION_REQUIREMENTS_NOT_MET', JText::_($extensionName), $versionMin), 'warning');
                }
            }
        }

        return $execution;
    }

    /**
     * Sets the default cache group depending on the device
     * This feature requires the plugin "Device Specific Content"
     * https://joomla-extensions.kubik-rubik.de/dsc-device-specific-content
     */
    private function setDefaultGroupDeviceSpecific()
    {
        if (file_exists(JPATH_PLUGINS . '/system/devicespecificcontent/devicespecificcontent.php')) {
            require_once JPATH_PLUGINS . '/system/devicespecificcontent/devicespecificcontent.php';

            if (PlgSystemDeviceSpecificContent::isMobile()) {
                $this->defaultGroup = $this->defaultGroup . '-mobile';
            } elseif (PlgSystemDeviceSpecificContent::isTablet()) {
                $this->defaultGroup = $this->defaultGroup . '-tablet';
            }
        }
    }

    /**
     * Checks the accepted encoding of the browser
     */
    private function checkEncodingGzip()
    {
        $acceptEncoding = '';

        if (!empty($_SERVER['HTTP_ACCEPT_ENCODING'])) {
            $acceptEncoding = $_SERVER['HTTP_ACCEPT_ENCODING'];
        }

        if (stripos($acceptEncoding, 'gzip') !== false) {
            $this->encodingGzip = true;
        }
    }

    /**
     * Loads cached files in the trigger onAfterInitialise
     */
    public function onAfterInitialise()
    {
        //S63 added 2 lines:
        $debugmode = JFactory::getApplication()->getCfg('debug');
        if ($debugmode == 1) {return false;}
        if (!$_SESSION['intro']) {return false;}
        
        if (count($this->app->getMessageQueue())) {
            return;
        }

        //S63$cleanCacheToken = $this->app->input->get('cleancache', false);
        $cleanCacheToken = $this->app->input->get('pagecache', false);

        // Purge process is requested - clean cache for loaded page and redirect to warm-up the cache
        if (!empty($cleanCacheToken) && $cleanCacheToken === $this->params->get('clean_cache_token', false)) {
            $uri = JUri::getInstance();
            $uri->delVar('cleancache');
            $urlCleaned = $uri->toString();
            $this->cleanCache($urlCleaned, $this->defaultGroup);
            $this->app->redirect($urlCleaned);
        }

        if (JFactory::getUser()->get('guest') && $this->app->input->getMethod() == 'GET') {
            $this->cache->setCaching(true);
        }

        $data = $this->loadStaticData();

        if (!empty($data)) {
            if ($this->encodingGzip) {
                $this->app->setHeader('Content-Encoding', 'gzip');
            }

            $this->app->setBody($data);
            echo $this->app->toString(false);
            $this->app->close();
        }
    }

    /**
     * Cleans cache of a specific page - normal and gzip version of the static file
     *
     * @param string $url
     * @param string $group
     */
    private function cleanCache($url, $group)
    {
        $this->cache->remove($url, $group . '-gzip');
        $this->cache->remove($url, $group);
    }

    /**
     * Loads the data from the stored static files - if data does not exist,
     * remove the newly created zero byte files from the cache check
     *
     * @return mixed
     */
    private function loadStaticData()
    {
        $group = $this->defaultGroup;

        if ($this->encodingGzip) {
            $group = $this->defaultGroup . '-gzip';
        }

        $data = $this->cache->get($this->cacheKey, $group);

        if (empty($data)) {
            $this->cache->remove($this->cacheKey, $group);
        }

        return $data;
    }

    /**
     * Creates the cache files in the trigger onAfterRender
     */
    public function onAfterRender()
    {
        if ($this->exclusionChecksPassed()) {
            if (count($this->app->getMessageQueue())) {
                return;
            }

            if (JFactory::getUser()->get('guest')) {
                $data = $this->app->getBody();

                // Only store data if not empty - this avoids storing of redirects, e.g. default language redirect
                if (!empty($data)) {
                    $minifyHtml = $this->params->get('minify_html', false);

                    if (!empty($minifyHtml)) {
                        $data = $this->minifyHtml($data, $minifyHtml == 2 ? true : false);
                    }

                    if (!empty($data)) {
                        // Save uncompressed file as fallback if browser doesn't support Gzip encoding
                        $this->cache->store($data, $this->cacheKey, $this->defaultGroup);

                        // Save also the Gzip variant as a file so that the data does not have to be
                        // compressed dynamically and can be transferred immediately
                        $data = gzencode($data, 4);
                        $this->cache->store($data, $this->cacheKey, $this->defaultGroup . '-gzip');
                    }
                }
            }
        }
    }

    /**
     * Checks whether the plugin should be executed on the loaded page
     *
     * @return bool
     * @throws Exception
     */
    private function exclusionChecksPassed()
    {
        $execute = true;

        if ($this->params->get('exclude_components')) {
            $this->excludeComponents($execute);
        }

        if ($this->params->get('exclude_urls')) {
            $this->excludeUrls($execute);
        }

        return $execute;
    }

    /**
     * Excludes or includes the execution of the plugin in specified components
     *
     * @param boolean $execute
     *
     * @throws Exception
     */
    private function excludeComponents(&$execute)
    {
        $option = $this->app->input->getWord('option');
        $excludeComponents = array_map('trim', explode("\n", $this->params->get('exclude_components')));
        $excludeComponentsToggle = (bool) $this->params->get('exclude_components_toggle', false);

        if (in_array($option, $excludeComponents)) {
            if (empty($excludeComponentsToggle)) {
                $execute = false;
            }

            return;
        }

        if (!empty($excludeComponentsToggle)) {
            $execute = false;
        }
    }

    /**
     * Excludes or includes the execution of the plugin in specified URLs
     *
     * @param boolean $execute
     */
    private function excludeUrls(&$execute)
    {
        $excludeUrls = array_map('trim', explode("\n", $this->params->get('exclude_urls')));
        $excludeUrlsToggle = (bool) $this->params->get('exclude_urls_toggle', false);

        if (in_array($this->cacheKey, $excludeUrls)) {
            if (!empty($excludeUrlsToggle)) {
                $execute = true;

                return;
            }

            $execute = false;
        }
    }

    /**
     * Minifies HTML code for the cached files to increase the loading time
     * Initially based on PHPWee PHP Minifier by SearchTurbine
     *
     * @param string $html
     * @param bool   $comments
     *
     * @return string
     */
    private function minifyHtml($html, $comments = false)
    {
        $doc = new DOMDocument();
        libxml_use_internal_errors(true);
        $doc->preserveWhiteSpace = false;
        $doc->loadHTML('<?xml encoding="utf-8" ?>' . $html);

        $domXpath = new DOMXPath($doc);

        if ($comments == false) {
            foreach ($domXpath->query('//comment()') as $comment) {
                if (strpos($comment->nodeValue, '[') === false) {
                    $comment->parentNode->removeChild($comment);
                }
            }
        }

        $doc->normalizeDocument();

        $nodesText = $domXpath->query('//text()');
        $skipHtml = array('style', 'pre', 'code', 'script', 'textarea');
        $skipTrim = array('a', 'img', 'span', 'strong', 'em', 'small', 'mark', 'del', 'ins', 'sub', 'sup', 'u', 'i');

        foreach ($nodesText as $node) {
            $nodePath = $node->getNodePath();

            if ($this->strposArray($nodePath, $skipHtml, '/') == false) {
                $node->nodeValue = preg_replace('@\s+@', ' ', $node->nodeValue);
            }

            if ($node->nodeType == 3 && $this->strposArray($nodePath, $skipTrim, '/') == false) {
                if (!empty($node->nextSibling->nodeName)) {
                    if ($this->strposArray($node->nextSibling->nodeName, $skipTrim) == true) {
                        continue;
                    }
                }

                if (!empty($node->previousSibling->nodeName)) {
                    if ($this->strposArray($node->previousSibling->nodeName, $skipTrim) == true) {
                        continue;
                    }
                }

                $node->nodeValue = trim($node->nodeValue);
            }
        }

        $doc->normalizeDocument();

        return '<!DOCTYPE html>' . "\n" . $doc->saveHTML($doc->documentElement);
    }

    /**
     * Small helper function for the PHP function strpos with the parameter $needle as an array
     *
     * @param string $haystack
     * @param array  $needle
     * @param string $prefix
     *
     * @return bool
     */
    private function strposArray($haystack, $needle, $prefix = '')
    {
        foreach ($needle as $needleValue) {
            if (strpos($haystack, $prefix . $needleValue) !== false) {
                return true;
            }
        }

        return false;
    }
}
